import os
import re
import sys

import pytest

import pybase64
from pybase64.__main__ import main


@pytest.fixture
def emptyfile(tmpdir):
    _file = os.path.join(tmpdir.strpath, "empty")
    with open(_file, "wb"):
        pass
    yield _file
    os.remove(_file)


@pytest.fixture
def hellofile(tmpdir):
    _file = os.path.join(tmpdir.strpath, "helloworld")
    with open(_file, "wb") as f:
        f.write(b"hello world !/?\n")
    yield _file
    os.remove(_file)


def idfn_test_help(args):
    if len(args) == 0:
        return "(empty)"
    return " ".join(args)


@pytest.mark.parametrize(
    "args",
    [
        [],
        ["-h"],
        ["benchmark", "-h"],
        ["encode", "-h"],
        ["decode", "-h"],
    ],
    ids=idfn_test_help,
)
def test_help(capsys, args):
    if len(args) == 2:
        usage = f"usage: pybase64 {args[0]} [-h]"
    else:
        usage = "usage: pybase64 [-h]"
    with pytest.raises(SystemExit) as exit_info:
        main(args)
    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out.startswith(usage)
    assert exit_info.value.code == 0


def test_version(capsys):
    with pytest.raises(SystemExit) as exit_info:
        main(["-V"])
    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out.startswith("pybase64 " + pybase64.__version__)
    assert exit_info.value.code == 0


def test_license(capsys):
    restr = "\n".join(
        x + "\n[=]+\n.*Copyright.*\n[=]+\n" for x in ["pybase64", "libbase64"]
    )
    regex = re.compile("^" + restr + "$", re.DOTALL)
    with pytest.raises(SystemExit) as exit_info:
        main(["--license"])
    captured = capsys.readouterr()
    assert captured.err == ""
    assert regex.match(captured.out)
    assert exit_info.value.code == 0


def test_benchmark(capsys, emptyfile):
    main(["benchmark", "-d", "0.005", emptyfile])
    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out != ""


@pytest.mark.parametrize(
    "args,expect",
    [
        ([], b"aGVsbG8gd29ybGQgIS8/Cg=="),
        (["-u"], b"aGVsbG8gd29ybGQgIS8_Cg=="),
        (["-a", ":,"], b"aGVsbG8gd29ybGQgIS8,Cg=="),
    ],
    ids=["0", "1", "2"],
)
def test_encode(capsysbinary, hellofile, args, expect):
    main(["encode"] + args + [hellofile])
    captured = capsysbinary.readouterr()
    assert captured.err == b""
    assert captured.out == expect


def test_encode_ouputfile(capsys, emptyfile, hellofile):
    main(["encode", "-o", hellofile, emptyfile])
    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out == ""
    with open(hellofile, "rb") as f:
        data = f.read()
    assert data == b""


@pytest.mark.parametrize(
    "args,b64string",
    [
        [[], b"aGVsbG8gd29ybGQgIS8/Cg=="],
        [["-u"], b"aGVsbG8gd29ybGQgIS8_Cg=="],
        [["-a", ":,"], b"aGVsbG8gd29ybGQgIS8,Cg=="],
        [["--no-validation"], b"aGVsbG8gd29yb GQgIS8/Cg==\n"],
    ],
    ids=["0", "1", "2", "3"],
)
def test_decode(capsysbinary, tmpdir, args, b64string):
    iname = os.path.join(tmpdir.strpath, "in")
    with open(iname, "wb") as f:
        f.write(b64string)
    main(["decode"] + args + [iname])
    captured = capsysbinary.readouterr()
    assert captured.err == b""
    assert captured.out == b"hello world !/?\n"


def test_subprocess():
    import subprocess

    process = subprocess.Popen(
        [sys.executable, "-m", "pybase64", "encode", "-"],
        bufsize=4096,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = process.communicate()
    process.wait()
    assert process.returncode == 0
    assert out == b""
    assert err == b""



from base64 import b64decode as builtin_decode
from base64 import b64encode as builtin_encode
from base64 import encodebytes as builtin_encodebytes
from binascii import Error as BinAsciiError
from typing import Any, Optional, Union

__all__ = [
    "_get_simd_path",
    "b64decode",
    "b64encode",
    "b64encode_as_string",
    "encodebytes",
]


_bytes_types = (bytes, bytearray)  # Types acceptable as binary data


def _get_simd_path() -> Optional[int]:
    return None


def _get_bytes(s: Any) -> Union[bytes, bytearray]:
    if isinstance(s, str):
        try:
            return s.encode("ascii")
        except UnicodeEncodeError:
            raise ValueError("string argument should contain only ASCII " "characters")
    if isinstance(s, _bytes_types):
        return s
    try:
        return memoryview(s).tobytes()
    except TypeError:
        raise TypeError(
            "argument should be a bytes-like object or ASCII "
            "string, not %r" % s.__class__.__name__
        )


def b64decode(s: Any, altchars: Any = None, validate: bool = False) -> bytes:
    """Decode bytes encoded with the standard Base64 alphabet.

    Argument ``s`` is a :term:`bytes-like object` or ASCII string to
    decode.

    Optional ``altchars`` must be a :term:`bytes-like object` or ASCII
    string of length 2 which specifies the alternative alphabet used instead
    of the '+' and '/' characters.

    If ``validate`` is ``False`` (the default), characters that are neither in
    the normal base-64 alphabet nor the alternative alphabet are discarded
    prior to the padding check.
    If ``validate`` is ``True``, these non-alphabet characters in the input
    result in a :exc:`binascii.Error`.

    The result is returned as a :class:`bytes` object.

    A :exc:`binascii.Error` is raised if ``s`` is incorrectly padded.
    """
    if validate:
        if len(s) % 4 != 0:
            raise BinAsciiError("Incorrect padding")
        s = _get_bytes(s)
        if altchars is not None:
            altchars = _get_bytes(altchars)
            assert len(altchars) == 2, repr(altchars)
            map = bytes.maketrans(altchars, b"+/")
            s = s.translate(map)
        try:
            result = builtin_decode(s, altchars, validate=False)
        except TypeError as e:
            raise BinAsciiError(str(e))

        # check length of result vs length of input
        padding = 0
        if len(s) > 1 and s[-2] in (b"=", 61):
            padding = padding + 1
        if len(s) > 0 and s[-1] in (b"=", 61):
            padding = padding + 1
        if 3 * (len(s) / 4) - padding != len(result):
            raise BinAsciiError("Non-base64 digit found")
        return result
    return builtin_decode(s, altchars, validate=False)


def b64decode_as_bytearray(
    s: Any, altchars: Any = None, validate: bool = False
) -> bytearray:
    """Decode bytes encoded with the standard Base64 alphabet.

    Argument ``s`` is a :term:`bytes-like object` or ASCII string to
    decode.

    Optional ``altchars`` must be a :term:`bytes-like object` or ASCII
    string of length 2 which specifies the alternative alphabet used instead
    of the '+' and '/' characters.

    If ``validate`` is ``False`` (the default), characters that are neither in
    the normal base-64 alphabet nor the alternative alphabet are discarded
    prior to the padding check.
    If ``validate`` is ``True``, these non-alphabet characters in the input
    result in a :exc:`binascii.Error`.

    The result is returned as a :class:`bytearray` object.

    A :exc:`binascii.Error` is raised if ``s`` is incorrectly padded.
    """
    return bytearray(b64decode(s, altchars=altchars, validate=validate))


def b64encode(s: Any, altchars: Any = None) -> bytes:
    """Encode bytes using the standard Base64 alphabet.

    Argument ``s`` is a :term:`bytes-like object` to encode.

    Optional ``altchars`` must be a byte string of length 2 which specifies
    an alternative alphabet for the '+' and '/' characters.  This allows an
    application to e.g. generate url or filesystem safe Base64 strings.

    The result is returned as a :class:`bytes` object.
    """
    if altchars is not None:
        altchars = _get_bytes(altchars)
        assert len(altchars) == 2, repr(altchars)
    return builtin_encode(s, altchars)


def b64encode_as_string(s: Any, altchars: Any = None) -> str:
    """Encode bytes using the standard Base64 alphabet.

    Argument ``s`` is a :term:`bytes-like object` to encode.

    Optional ``altchars`` must be a byte string of length 2 which specifies
    an alternative alphabet for the '+' and '/' characters.  This allows an
    application to e.g. generate url or filesystem safe Base64 strings.

    The result is returned as a :class:`str` object.
    """
    return b64encode(s, altchars).decode("ascii")


def encodebytes(s: Any) -> bytes:
    """Encode bytes into a bytes object with newlines (b'\n') inserted after
    every 76 bytes of output, and ensuring that there is a trailing newline,
    as per :rfc:`2045` (MIME).

    Argument ``s`` is a :term:`bytes-like object` to encode.

    The result is returned as a :class:`bytes` object.
    """
    return builtin_encodebytes(s)



import builtins
import glob
import os
import runpy
import sys
from codecs import open
from distutils import log
from distutils.cmd import Command
from distutils.command.build_ext import build_ext
from distutils.dep_util import newer_group
from distutils.errors import DistutilsSetupError
from os import environ, path

from setuptools import Extension, find_packages, setup

# Not to try loading things from the main module during setup
builtins.__PYBASE64_SETUP__ = True

from pybase64.distutils.ccompilercapabilities import CCompilerCapabilities  # noqa: E402

if sys.version_info[:2] < (3, 6):
    raise RuntimeError("Python version >= 3.6 required.")

here = path.abspath(path.dirname(__file__))

# Get version
version = runpy.run_path(path.join(here, "pybase64", "_version.py"))["__version__"]

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

# Generate license text
with open(path.join(here, "pybase64/_license.py"), "wt") as f:
    f.write('_license = """')
    f.write("pybase64\n")
    f.write("=" * 79 + "\n")
    with open(path.join(here, "LICENSE")) as license:
        f.write(license.read())
    f.write("=" * 79 + "\n\n")
    f.write("libbase64\n")
    f.write("=" * 79 + "\n")
    with open(path.join(here, "base64/LICENSE")) as license:
        f.write(license.read())
    f.write("=" * 74)
    f.write('""" \\\n')
    f.write('    + "====="\n')

pybase64_ext = Extension(
    "pybase64._pybase64",
    [
        "pybase64/_pybase64.c",
        "pybase64/_pybase64_get_simd_flags.c",
        "base64/lib/lib.c",
        "base64/lib/codec_choose.c",
        "base64/lib/arch/generic/codec.c",
        "base64/lib/tables/tables.c",
    ],
    include_dirs=["base64/include/", "base64/lib/"],
    define_macros=[("BASE64_STATIC_DEFINE", "1")],
)

pybase64_ext.optional = environ.get("CIBUILDWHEEL", "0") != "1"

pybase64_ext.sources_ssse3 = ["base64/lib/arch/ssse3/codec.c"]
pybase64_ext.sources_sse41 = ["base64/lib/arch/sse41/codec.c"]
pybase64_ext.sources_sse42 = ["base64/lib/arch/sse42/codec.c"]
pybase64_ext.sources_avx = ["base64/lib/arch/avx/codec.c"]
pybase64_ext.sources_avx2 = ["base64/lib/arch/avx2/codec.c"]
pybase64_ext.sources_neon32 = ["base64/lib/arch/neon32/codec.c"]
pybase64_ext.sources_neon64 = ["base64/lib/arch/neon64/codec.c"]


def pybase64_write_config(capabilities):
    log.info("creating 'base64/lib/config.h'")
    with open(path.join(here, "base64/lib/config.h"), mode="wt") as f:
        f.write(
            "\n#define HAVE_SSSE3                 %i"
            % capabilities.has(CCompilerCapabilities.SIMD_SSSE3)
        )
        f.write(
            "\n#define HAVE_SSE41                 %i"
            % capabilities.has(CCompilerCapabilities.SIMD_SSE41)
        )
        f.write(
            "\n#define HAVE_SSE42                 %i"
            % capabilities.has(CCompilerCapabilities.SIMD_SSE42)
        )
        f.write(
            "\n#define HAVE_AVX                   %i"
            % capabilities.has(CCompilerCapabilities.SIMD_AVX)
        )
        f.write(
            "\n#define HAVE_AVX2                  %i"
            % capabilities.has(CCompilerCapabilities.SIMD_AVX2)
        )
        f.write(
            "\n#define HAVE_NEON32                %i"
            % capabilities.has(CCompilerCapabilities.SIMD_NEON32)
        )
        f.write(
            "\n#define HAVE_NEON64                %i"
            % capabilities.has(CCompilerCapabilities.SIMD_NEON64)
        )
        f.write(
            "\n#define HAVE_FAST_UNALIGNED_ACCESS %i"
            % capabilities.has(CCompilerCapabilities.SIMD_SSSE3)
        )  # on x86
        f.write("\n")


class pybase64_build_ext(build_ext):
    def build_extension(self, ext):
        sources = ext.sources
        if sources is None or not isinstance(sources, (list, tuple)):
            raise DistutilsSetupError(
                "in 'ext_modules' option (extension '%s'), "
                "'sources' must be present and must be "
                "a list of source filenames" % ext.name
            )
        sources = list(sources)

        simd_sources = {
            CCompilerCapabilities.SIMD_SSSE3: ext.sources_ssse3,
            CCompilerCapabilities.SIMD_SSE41: ext.sources_sse41,
            CCompilerCapabilities.SIMD_SSE42: ext.sources_sse42,
            CCompilerCapabilities.SIMD_AVX: ext.sources_avx,
            CCompilerCapabilities.SIMD_AVX2: ext.sources_avx2,
            CCompilerCapabilities.SIMD_NEON32: ext.sources_neon32,
            CCompilerCapabilities.SIMD_NEON64: ext.sources_neon64,
        }

        ext_path = self.get_ext_fullpath(ext.name)
        depends = sources + ext.depends

        try:
            simd_sources_values = simd_sources.itervalues()
        except AttributeError:
            simd_sources_values = simd_sources.values()
        for add_sources in simd_sources_values:
            depends = depends + add_sources

        if not (self.force or newer_group(depends, ext_path, "newer")):
            log.debug("skipping '%s' extension (up-to-date)", ext.name)
            return
        else:
            log.info("building '%s' extension", ext.name)

        capabilities = CCompilerCapabilities(self.compiler)
        pybase64_write_config(capabilities)

        objects = []

        for simd_opt in (
            CCompilerCapabilities.SIMD_SSSE3,
            CCompilerCapabilities.SIMD_SSE41,
            CCompilerCapabilities.SIMD_SSE42,
            CCompilerCapabilities.SIMD_AVX,
            CCompilerCapabilities.SIMD_AVX2,
            CCompilerCapabilities.SIMD_NEON32,
            CCompilerCapabilities.SIMD_NEON64,
        ):
            if len(simd_sources[simd_opt]) == 0:
                continue
            if capabilities.has(simd_opt):
                objects = objects + self.compiler.compile(
                    simd_sources[simd_opt],
                    output_dir=self.build_temp,
                    include_dirs=ext.include_dirs,
                    debug=self.debug,
                    extra_postargs=capabilities.flags(simd_opt),
                    depends=ext.depends,
                    macros=ext.define_macros,
                )
            else:
                sources = sources + simd_sources[simd_opt]

        objects = objects + self.compiler.compile(
            sources,
            output_dir=self.build_temp,
            include_dirs=ext.include_dirs,
            debug=self.debug,
            extra_postargs=[],
            depends=ext.depends,
            macros=ext.define_macros,
        )

        # Detect target language, if not provided
        language = ext.language or self.compiler.detect_language(sources)

        self.compiler.link_shared_object(
            objects,
            ext_path,
            libraries=self.get_libraries(ext),
            library_dirs=ext.library_dirs,
            runtime_library_dirs=ext.runtime_library_dirs,
            export_symbols=self.get_export_symbols(ext),
            debug=self.debug,
            build_temp=self.build_temp,
            target_lang=language,
        )


# Let's define a class to clean in-place built extensions
class CleanExtensionCommand(Command):
    """A custom command to clean all in-place built C extensions."""

    description = "clean all in-place built C extensions"
    user_options = []

    def initialize_options(self):
        """Set default values for options."""

    def finalize_options(self):
        """Post-process options."""

    def run(self):
        """Run command."""
        for ext in ["*.so", "*.pyd"]:
            for file in glob.glob("./pybase64/" + ext):
                log.info("removing '%s'", file)
                if self.dry_run:
                    continue
                os.remove(file)


# Get the C code
exts = [pybase64_ext]


setup(
    name="pybase64",
    cmdclass={
        "build_ext": pybase64_build_ext,
        "clean_ext": CleanExtensionCommand,
    },
    ext_modules=exts,
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,
    description="Fast Base64 encoding/decoding",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # The project's main homepage.
    url="https://github.com/mayeut/pybase64",
    project_urls={
        "Source": "https://github.com/mayeut/pybase64",
        "Tracker": "https://github.com/mayeut/pybase64/issues",
        "Documentation": "https://pybase64.readthedocs.io/en/stable",
    },
    # Author details
    author="Matthieu Darbois",
    # author_email = 'mayeut@users.noreply.github.com',
    # Choose your license
    license="BSD-2-Clause",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: BSD License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    # Supported python versions
    python_requires=">=3.6",
    # What does your project relate to?
    keywords="base64",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        "test": ["pytest>=5.0.0"],
    },
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        "console_scripts": [
            "pybase64=pybase64.__main__:main",
        ],
    },
)



import os
import shutil
import sys
import tempfile
from contextlib import contextmanager
from distutils import log
from distutils.errors import CompileError, LinkError

__all__ = ["CCompilerCapabilities"]


@contextmanager
def chdir(path):
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)


@contextmanager
def mkdtemp(suffix):
    path = tempfile.mkdtemp(suffix)
    try:
        with chdir(path):
            yield path
    finally:
        shutil.rmtree(path)


@contextmanager
def output(is_quiet):
    if is_quiet:  # pragma: no branch
        devnull = open(os.devnull, "w")
        oldstderr = os.dup(sys.stderr.fileno())
        oldstdout = os.dup(sys.stdout.fileno())
        os.dup2(devnull.fileno(), sys.stderr.fileno())
        os.dup2(devnull.fileno(), sys.stdout.fileno())
    try:
        yield
    finally:
        if is_quiet:  # pragma: no branch
            os.dup2(oldstderr, sys.stderr.fileno())
            os.dup2(oldstdout, sys.stdout.fileno())
            devnull.close()


class CCompilerCapabilities:
    SIMD_SSSE3 = 0
    SIMD_SSE41 = 1
    SIMD_SSE42 = 2
    SIMD_AVX = 3
    SIMD_AVX2 = 4
    SIMD_NEON32 = 5
    SIMD_NEON64 = 6

    def __init__(self, compiler):
        self.__capabilities = {}
        self.__cflags = []
        if compiler.compiler_type == "msvc":
            self.__cflags = ["/WX", "/Od"]  # pragma: no cover
        else:
            self.__cflags = ["-O0"]
        self.__get_capabilities(compiler)

    def __has_simd_support(self, compiler, flags, define, include, content):
        quiet = True

        with mkdtemp("pybase64simdtest") as dname:
            fname = os.path.join(dname, "simd.c")
            with open(fname, "w") as f:
                f.write("""#include <%s>\n""" % include)
                f.write(
                    """\
int main (int argc, char **argv) {
    %s
}
"""
                    % content
                )
            with output(quiet):
                for flag in flags:
                    lflags = []
                    if not len(flag) == 0:
                        lflags = [flag]
                    try:
                        objects = compiler.compile(
                            ["simd.c"],
                            output_dir=dname,
                            extra_postargs=lflags + self.__cflags,
                        )
                    except CompileError:
                        continue
                    try:
                        compiler.link_shared_lib(objects, "a.out", output_dir=dname)
                    except (LinkError, TypeError):  # pragma: no cover
                        continue  # pragma: no cover
                    return {"support": True, "flags": lflags + define}
                return {"support": False, "flags": []}

    def __get_capabilities(self, compiler):
        log.info("getting compiler simd support")
        self.__capabilities[CCompilerCapabilities.SIMD_SSSE3] = self.__has_simd_support(
            compiler,
            ["", "-mssse3"],
            ["-D__SSSE3__"],
            "tmmintrin.h",
            "__m128i t = _mm_loadu_si128((const __m128i*)argv[0]);"
            "t = _mm_shuffle_epi8(t, t);"
            "return _mm_cvtsi128_si32(t);",
        )
        log.info(
            "SSSE3:  %s"
            % str(self.__capabilities[CCompilerCapabilities.SIMD_SSSE3]["support"])
        )
        self.__capabilities[CCompilerCapabilities.SIMD_SSE41] = self.__has_simd_support(
            compiler,
            ["", "-msse4.1"],
            ["-D__SSE4_1__"],
            "smmintrin.h",
            "__m128i t = _mm_loadu_si128((const __m128i*)argv[0]);"
            "t = _mm_mpsadbw_epu8(t, t, 1);"
            "return _mm_cvtsi128_si32(t);",
        )
        log.info(
            "SSE41:  %s"
            % str(self.__capabilities[CCompilerCapabilities.SIMD_SSE41]["support"])
        )
        self.__capabilities[CCompilerCapabilities.SIMD_SSE42] = self.__has_simd_support(
            compiler,
            ["", "-msse4.2"],
            ["-D__SSE4_2__"],
            "nmmintrin.h",
            "__m128i t = _mm_loadu_si128((const __m128i*)argv[0]);"
            "return _mm_cmpistra(t, t, 0);",
        )
        log.info(
            "SSE42:  %s"
            % str(self.__capabilities[CCompilerCapabilities.SIMD_SSE42]["support"])
        )
        self.__capabilities[CCompilerCapabilities.SIMD_AVX] = self.__has_simd_support(
            compiler,
            ["", "-mavx", "/arch:AVX"],
            ["-D__AVX__"],
            "immintrin.h",
            "__m256i y = _mm256_loadu_si256((const __m256i*)argv[0]);"
            "return _mm256_testz_si256(y, y);",
        )
        log.info(
            "AVX:    %s"
            % str(self.__capabilities[CCompilerCapabilities.SIMD_AVX]["support"])
        )
        self.__capabilities[CCompilerCapabilities.SIMD_AVX2] = self.__has_simd_support(
            compiler,
            ["", "-mavx2", "/arch:AVX2"],
            ["-D__AVX2__"],
            "immintrin.h",
            "__m256i y = _mm256_loadu_si256((const __m256i*)argv[0]);"
            "y = _mm256_i32gather_epi32((int const*)argv[1], y, 2);"
            "return _mm_cvtsi128_si32(_mm256_castsi256_si128(y));",
        )
        log.info(
            "AVX2:   %s"
            % str(self.__capabilities[CCompilerCapabilities.SIMD_AVX2]["support"])
        )
        self.__capabilities[
            CCompilerCapabilities.SIMD_NEON64
        ] = self.__has_simd_support(
            compiler,
            [""],
            [],
            "arm_neon.h",
            "uint8x16_t t = vdupq_n_u8(1);"
            "uint8x16x4_t t4 = {"
            "    .val[0]=t, .val[1]=t, .val[2]=t, .val[3]=t};"
            "uint8x16_t o = vqtbx4q_u8(t, t4, t);"
            "return vgetq_lane_s32(vreinterpretq_s32_u8(o), 0);",
        )
        log.info(
            "NEON64: %s"
            % str(self.__capabilities[CCompilerCapabilities.SIMD_NEON64]["support"])
        )
        self.__capabilities[
            CCompilerCapabilities.SIMD_NEON32
        ] = self.__has_simd_support(
            compiler,
            [""],
            [],
            "arm_neon.h",
            "uint8x16_t t = vdupq_n_u8(1);"
            "return vgetq_lane_s32(vreinterpretq_s32_u8(t), 0);",
        )
        self.__capabilities[CCompilerCapabilities.SIMD_NEON32][
            "support"
        ] &= not self.__capabilities[CCompilerCapabilities.SIMD_NEON64]["support"]
        log.info(
            "NEON32: %s"
            % str(self.__capabilities[CCompilerCapabilities.SIMD_NEON32]["support"])
        )

    def has(self, what):
        if what not in self.__capabilities:
            return False
        return self.__capabilities[what]["support"]

    def flags(self, what):
        if not self.has(what):  # pragma: no branch
            return self.__capabilities[666]["flags"]  # pragma: no cover
        return self.__capabilities[what]["flags"]
